package sim.app.species.ANN;

import ec.util.MersenneTwisterFast;
import java.util.StringTokenizer;

class AxonGene extends Gene{

	private double weight;
	private int startID, endID;
	private boolean disabled;

	public AxonGene(int id, int start, int end, double w){

		super(id);

		this.startID = start;
		this.endID = end;
		this.weight = w;
		this.disabled = false;

	}

	public AxonGene(int id, int start, int end, double w, boolean dis){

		this(id, start, end, w);

		this.disabled = dis;

	}

	public AxonGene(int id, int start, int end, MersenneTwisterFast random){

		this(id, start, end, 0.0);

		this.weight = random.nextDouble() * Axon.getMaxWeight();

		if(random.nextBoolean()) this.weight *= -1.0;

	}

	public static AxonGene makeGeneFromString(String str){

		StringTokenizer tokenizer = new StringTokenizer(str);
		int i, sid, eid;
		double w;
		boolean dis;

		tokenizer.nextToken();
		i = Integer.parseInt(tokenizer.nextToken());

		tokenizer.nextToken();
		sid = Integer.parseInt(tokenizer.nextToken());

		tokenizer.nextToken();
		eid = Integer.parseInt(tokenizer.nextToken());

		tokenizer.nextToken();
		w = Double.parseDouble(tokenizer.nextToken());

		tokenizer.nextToken();
		dis = Boolean.getBoolean(tokenizer.nextToken());

		return new AxonGene(i, sid, eid, w, dis);

	}


	public void setDisabled(boolean b){
		this.disabled = b;
	}

	public boolean getDisabled(){
		return this.disabled;
	}

	public void addWeight(double delta){
		this.weight += delta;
	}

	public double getWeight(){
		return this.weight;
	}

	public int getStartID(){
		return this.startID;
	}

	public int getEndID(){
		return this.endID;
	}

	public String toString(){

		return "ID: " + this.getID() + " StartID: " + this.startID + " EndID: " + this.endID + " Weight: " + this.weight + " Disabled: " + this.disabled;

	}

}
