package sim.app.species;

import sim.engine.*;
import sim.portrayal.*;
import sim.util.*;
import java.util.*;

public class Bullet extends Entity{

	private static int DEFAULT_INIT_ENERGY = 200;

	private static boolean DEFAULT_SHOW_LOOK = false;
	private static double DEFAULT_DIAMETER = 75.0;
	private static double DEFAULT_SENSE_DIST = 75.0;
	private static double DEFAULT_SPEED = 10.0;
	private static boolean[] DEFAULT_LOOK = new boolean[Entity.NUM_BITS_IN_LOOK];
	private static int DEFAULT_LIFE_SPAN = 0;

	public Bullet(MutableDouble2D initPos, MutableDouble2D initHead){

		super(initPos, initHead);

		this.setDiameter(Bullet.DEFAULT_DIAMETER);
		this.setShowLook(Bullet.DEFAULT_SHOW_LOOK);
		this.setSenseDistance(Bullet.DEFAULT_SENSE_DIST);
		this.setSpeed(Bullet.DEFAULT_SPEED);
		this.setLook(Bullet.DEFAULT_LOOK);
		this.setLifeSpan(Bullet.DEFAULT_LIFE_SPAN);

		this.resetColor();

		this.setEnergy(Bullet.DEFAULT_INIT_ENERGY);

	}

	public Bullet(){ super(); }

	public static void setDefaults(boolean showLook, double diameter, double speed, boolean[] look, double senseDist, int lifeSpan){

		Bullet.DEFAULT_SHOW_LOOK = showLook;
		Bullet.DEFAULT_DIAMETER = diameter;
		Bullet.DEFAULT_SPEED = speed;
		Bullet.DEFAULT_LOOK = look;
		Bullet.DEFAULT_SENSE_DIST = senseDist;
		Bullet.DEFAULT_LIFE_SPAN = lifeSpan;

	}

	public static void setInitialEnergy(int n){

		Bullet.DEFAULT_INIT_ENERGY = n;

	}

	public void step(SimState state){

		SpeciesSim sim = (SpeciesSim)state;

		this.interactFood(sim);

		if(this.getEnergy() <= 0){

			sim.removeEntityFromSim(this, sim.bullets);
			return;

		}else{
			this.addEnergy(-1);
		}

		/* when bullets hit walls, they die!
		 **/
		if(this.hittingWall(sim.fieldWidth, sim.fieldHeight)){

			sim.removeEntityFromSim(this, sim.bullets);
			return;

		}

		/* update the position in the Continuous2D container
		 * also, update the position within the agent
		 **/
		this.updatePosition();
		sim.bullets.setObjectLocation(this, new Double2D(this.position));

	}

	private void interactFood(SpeciesSim sim){

		Bag nearbyCreatures = sim.creatures.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		Creature creature;

		for(int i = 0; i < nearbyCreatures.numObjs; i++){

			creature = (Creature)nearbyCreatures.objs[i];

			// check to see if the bullet is touching the creature.  if it is, then kill the creature and continue to next one
			if(this.isTouching(creature)){

				sim.removeEntityFromSim(creature, sim.creatures);

			}

		}

	}

}
