package sim.app.species;

import sim.engine.*;
import sim.portrayal.*;
import sim.util.*;
import java.util.*;
import sim.app.species.ANN.*;
import java.awt.Color;

public class Creature extends Entity{

	private static SpeciesSim sSim;

	private static int idCount = 0;
	private static final int NUM_ACTIONS = 3;

	private static boolean DEFAULT_SHOW_LOOK = false;
	private static double DEFAULT_DIAMETER = 75.0;
	private static double DEFAULT_SENSE_DIST = 75.0;
	private static double DEFAULT_SPEED = 10.0;
	private static boolean[] DEFAULT_LOOK = new boolean[Entity.NUM_BITS_IN_LOOK];
	private static int DEFAULT_LIFE_SPAN = 1000;

	private static int SHOW_MATING_ALERT_TIME = 10;
	private static boolean SHOW_SENSOR = false;
	private static int VISION_DIVISIONS = 3;
	private static double SHARP_ANGLE_TURN = 0.05;
	private static double WIDE_ANGLE_TURN = 0.025;
	private static double DELTA_SPEED = 1.0;
	private static double LOOK_MUT_RATE = 0.05;
	private static int MATING_ENERGY_THRESHOLD = 10000;
	private static int ASEXUAL_MATE_THRESHOLD = 20000;
	private static int BABY_STARTING_ENERGY = 5000;

	private static Color[] speciesColors = new Color[]{};
	private static String matingRestriction = "";
	private static boolean canSeeOtherCreatures = false;
	private static String inputToBrainType = "";

	protected Brain brain;
	private int id, speciesNum;
	protected int justMated;
	private int movingFoodEaten, stationaryFoodEaten;
	private int sexualReproductions, asexualReproductions;
	protected boolean isBestMovingFoodEater, isBestStationaryFoodEater;

	public Creature(MutableDouble2D initPos, MutableDouble2D initHead, Brain brain){

		super(initPos, initHead);

		this.id = Creature.idCount++;
		this.movingFoodEaten = 0;
		this.stationaryFoodEaten = 0;
		this.asexualReproductions = 0;
		this.sexualReproductions = 0;
		this.justMated = 0;

		this.setDiameter(Creature.DEFAULT_DIAMETER);
		this.setShowLook(Creature.DEFAULT_SHOW_LOOK);
		this.setSenseDistance(Creature.DEFAULT_SENSE_DIST);
		this.setSpeed(Creature.DEFAULT_SPEED);
		this.setLook(Creature.DEFAULT_LOOK);
		this.setLifeSpan(Creature.DEFAULT_LIFE_SPAN);
		this.setDrawSensor(Creature.SHOW_SENSOR);

		this.resetColor();

		// WILL CHANGE!!!!!!
		this.setEnergy(2500);
		
		this.brain = brain;

		this.setDrawHeading(true);

		this.isBestMovingFoodEater = false;
		this.isBestStationaryFoodEater = false;

	}

	public Creature(){ super(); }

	public static void setDefaults(boolean showLook, double diameter, double speed, boolean[] look, double senseDist, int lifeSpan){

		Creature.DEFAULT_SHOW_LOOK = showLook;
		Creature.DEFAULT_DIAMETER = diameter;
		Creature.DEFAULT_SPEED = speed;
		Creature.DEFAULT_LOOK = look;
		Creature.DEFAULT_SENSE_DIST = senseDist;
		Creature.DEFAULT_LIFE_SPAN = lifeSpan;

	}

	public static void setSpeciesColors(Color[] arr){
		Creature.speciesColors = arr;
	}

	public static void setCanSeeOtherCreatures(boolean b){
		Creature.canSeeOtherCreatures = b;
	}

	public static void setSpeciesSimReference(SpeciesSim sim){
		Creature.sSim = sim;
	}

	public static void setInputToBrainType(String str){
		Creature.inputToBrainType = str;
	}

	/* simply updates the agent's position based off of its heading and current position
	 **/
	public void step(SimState state){

		SpeciesSim sim = (SpeciesSim)state;

		// kill the creature if it has no energy or it dies of old age
		if(this.getEnergy() <= 0 || this.getAge() > this.getLifeSpan()){
			sim.removeEntityFromSim(this, sim.creatures);
			return;
		}

		if(this.justMated > 0) this.justMated--;

		// decrease the energy based on the speed
		if(this.getSpeed() == 0.0){
			this.addEnergy(-1);
		}else{
			this.addEnergy(-1);
		}
		this.incrementAge();

		double[] outp = this.brain.activate(this.inputToBrain(sim));

		if(outp[0] == 1.0) this.heading.rotate(-Creature.SHARP_ANGLE_TURN);
		if(outp[1] == 1.0) this.heading.rotate(Creature.SHARP_ANGLE_TURN);

		//if(outp[2] == 1.0) this.heading.rotate(-Creature.WIDE_ANGLE_TURN);
		//if(outp[3] == 1.0) this.heading.rotate(Creature.WIDE_ANGLE_TURN);

		if(outp[2] == 1.0) this.addSpeed(Creature.DEFAULT_SPEED);
		else               this.addSpeed(0.0);
		//if(outp[3] == 1.0) this.addSpeed(Creature.DELTA_SPEED);

		this.interactFood(sim);
		this.mateWithNeighbors(sim);
		this.asexualReproduction(sim);

		// updates the heading of the entity if it is hitting a wall
		this.updateHeadingHitWall(sim.fieldWidth, sim.fieldHeight);

		/* update the position in the Continuous2D container
		 * also, update the position within the agent
		 **/
		this.updatePosition();
		sim.creatures.setObjectLocation(this, new Double2D(this.position));

	}

	// returns a 
	private double[] inputToBrain(SpeciesSim sim){

		if(Creature.inputToBrainType.equals("EXPLICIT")){
			return this.explicitInputToBrain(sim);
		}else if(Creature.inputToBrainType.equals("EXPLICIT_SIMPLE")){
			return this.explicitInputToBrainSimple(sim);
		}else if(Creature.inputToBrainType.equals("IMPLICIT")){
			return this.implicitInputToBrain(sim);
		}

		return null;

	}

	private double[] explicitInputToBrainSimple(SpeciesSim sim){

		Bag nearbyEntities;
		double[] input = new double[Creature.numInputsToBrain()];
		boolean[][] entityVisible = new boolean[3][Creature.numVisionDivisions() + 1];

		if(Creature.canSeeOtherCreatures){
			nearbyEntities = sim.creatures.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
			this.explicitLookAtEntities(nearbyEntities, entityVisible[0]);
		}

		nearbyEntities = sim.movingFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[1]);

		nearbyEntities = sim.stationaryFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[2]);

		for(int i = 0; i < entityVisible.length; i++){
			for(int j = 0; j < entityVisible[i].length; j++){
				if(entityVisible[i][j]){
					input[(i*entityVisible[i].length) + j] = 5.0;
				}else{
					input[(i*entityVisible[i].length) + j] = -5.0;
				}
			}
		}

		return input;

	}
	
	private double[] explicitInputToBrain(SpeciesSim sim){

		Bag nearbyEntities;
		double[] input = new double[Creature.numInputsToBrain()];
		boolean[][] entityVisible = new boolean[6][Creature.numVisionDivisions() + 1];

		if(Creature.canSeeOtherCreatures){
			nearbyEntities = sim.creatures.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
			this.explicitLookAtEntities(nearbyEntities, entityVisible[0]);
		}

		nearbyEntities = sim.movingFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[1]);

		nearbyEntities = sim.predators.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[2]);

		nearbyEntities = sim.stationaryFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[3]);

		nearbyEntities = sim.turrets.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[4]);

		nearbyEntities = sim.bullets.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.explicitLookAtEntities(nearbyEntities, entityVisible[5]);

		for(int i = 0; i < entityVisible.length; i++){
			for(int j = 0; j < entityVisible[i].length; j++){
				if(entityVisible[i][j]){
					input[(i*entityVisible[i].length) + j] = 5.0;
				}else{
					input[(i*entityVisible[i].length) + j] = -5.0;
				}
			}
		}

		return input;

	}

	private void explicitLookAtEntities(Bag nearbyEntities, boolean[] isVisible){

		MutableDouble2D vectorToEntity = new MutableDouble2D();
		Entity entity;
		double angle, perpDot, dist;
		double angleDivisions = Math.PI/Creature.numVisionDivisions();

		for(int i = 0; i < nearbyEntities.numObjs; i++){

			entity = (Entity)nearbyEntities.objs[i];

			if(!entity.isAlive()) continue;

			vectorToEntity.subtract(entity.position, this.position);
			dist = vectorToEntity.length();

			if(dist > 0.0 && dist - entity.getRadius() <= this.senseDistance()){

				vectorToEntity.normalize();

				angle = this.heading.dot(vectorToEntity);

				// check to see if the entity is in the front
				if(angle > 0.0){

					angle = Math.acos(angle);

					perpDot = this.heading.perpDot(vectorToEntity);

					if(perpDot > 0.0){
						angle = (Math.PI/2.0) - angle;
					}else{
						angle = (Math.PI/2.0) + angle;
					}

					for(int j = 0; j < isVisible.length - 1; j++){

						if(angle <= angleDivisions * (j+1)){
							isVisible[j] = true;
							break;
						}

					}

				// the entity can see behind it also
				}else{

					isVisible[isVisible.length-1] = true;

				}

			}

		}

	}

	private double[] implicitInputToBrain(SpeciesSim sim){

		Bag nearbyEntities;
		double[] input = new double[Creature.numInputsToBrain()];
		double[] minDists = new double[Creature.numVisionDivisions() + 1];
		Entity[] minDistEntities = new Entity[Creature.numVisionDivisions() + 1];

		Arrays.fill(minDists, Double.MAX_VALUE);

		if(Creature.canSeeOtherCreatures){
			nearbyEntities = sim.creatures.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
			this.implicitLookAtEntities(nearbyEntities, minDists, minDistEntities);
		}

		nearbyEntities = sim.movingFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.implicitLookAtEntities(nearbyEntities, minDists, minDistEntities);

		nearbyEntities = sim.predators.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.implicitLookAtEntities(nearbyEntities, minDists, minDistEntities);

		nearbyEntities = sim.stationaryFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.implicitLookAtEntities(nearbyEntities, minDists, minDistEntities);

		nearbyEntities = sim.turrets.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.implicitLookAtEntities(nearbyEntities, minDists, minDistEntities);

		nearbyEntities = sim.bullets.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		this.implicitLookAtEntities(nearbyEntities, minDists, minDistEntities);

		for(int i = 0; i < minDistEntities.length; i++){

			if(minDistEntities[i] == null){
				for(int j = 0; j < Entity.NUM_BITS_IN_LOOK; j++){
					input[i*Entity.NUM_BITS_IN_LOOK + j] = -5.0;
				}
			}else{
				boolean[] entityLook = minDistEntities[i].getLook();
				for(int j = 0; j < Entity.NUM_BITS_IN_LOOK; j++){
					if(entityLook[j]){
						input[i*Entity.NUM_BITS_IN_LOOK + j] = 5.0;
					}else{
						input[i*Entity.NUM_BITS_IN_LOOK + j] = 0.0;
					}
				}
			}

		}

		return input;

	}

	private void implicitLookAtEntities(Bag nearbyEntities, double[] minDists, Entity[] minDistEntities){

		MutableDouble2D vectorToEntity = new MutableDouble2D();
		Entity entity;
		double angle, perpDot, dist;
		double angleDivisions = Math.PI/Creature.numVisionDivisions();

		for(int i = 0; i < nearbyEntities.numObjs; i++){

			entity = (Entity)nearbyEntities.objs[i];

			if(!entity.isAlive()) continue;

			vectorToEntity.subtract(entity.position, this.position);
			dist = vectorToEntity.length();

			if(dist > 0.0 && dist - entity.getRadius() <= this.senseDistance()){

				vectorToEntity.normalize();

				angle = this.heading.dot(vectorToEntity);

				// check to see if the entity is in the front
				if(angle > 0.0){

					angle = Math.acos(angle);

					perpDot = this.heading.perpDot(vectorToEntity);

					if(perpDot > 0.0){
						angle = (Math.PI/2.0) - angle;
					}else{
						angle = (Math.PI/2.0) + angle;
					}

					for(int j = 0; j < minDists.length - 1; j++){

						if(angle <= angleDivisions * (j+1)){
							if(minDists[j] > dist){
								minDists[j] = dist;
								minDistEntities[j] = entity;
							}
							break;
						}

					}

				// the entity can see behind it also
				}else if(minDists[minDists.length-1] > dist){

					minDists[minDists.length-1] = dist;
					minDistEntities[minDists.length-1] = entity;

				}

			}

		}

	}


	// eat nearby food
	private void interactFood(SpeciesSim sim){

		Bag nearbyEntities = sim.movingFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		MovingFood movingFood;
		StationaryFood stationaryFood;

		for(int i = 0; i < nearbyEntities.numObjs; i++){

			movingFood = (MovingFood)nearbyEntities.objs[i];

			if(!movingFood.isAlive()) continue;

			// check to see if the creature is touching the food.  if it is, then eat the food and continue to next one
			if(this.isTouching(movingFood)){

				// get the energy from the food
				this.addEnergy(movingFood.getEnergy());
				this.movingFoodEaten++;
				sim.removeEntityFromSim(movingFood, sim.movingFood);

				continue;
			}

		}

		nearbyEntities = sim.stationaryFood.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());

		for(int i = 0; i < nearbyEntities.numObjs; i++){

			stationaryFood = (StationaryFood)nearbyEntities.objs[i];

			if(!stationaryFood.isAlive()) continue;

			// check to see if the creature is touching the food.  if it is, then eat the food and continue to next one
			if(this.isTouching(stationaryFood)){

				// get the energy from the food
				this.addEnergy(stationaryFood.getEnergy());
				this.stationaryFoodEaten++;
				sim.removeEntityFromSim(stationaryFood, sim.stationaryFood);

				continue;
			}

		}

	}

	private void mateWithNeighbors(SpeciesSim sim){

		Bag nearbyEntities = sim.creatures.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		Creature creature;

		for(int i = 0; i < nearbyEntities.numObjs; i++){

			// this creature does not have enough energy to mate, so it dies
			if(this.getEnergy() < Creature.MATING_ENERGY_THRESHOLD) break;

			creature = (Creature)nearbyEntities.objs[i];

			if(!creature.isAlive() || creature.getEnergy() < Creature.MATING_ENERGY_THRESHOLD || this == creature) continue;

			/* check to see if the creature is touching the creature and the energy thresholds are
			 *   high enough so they can mate
			 **/
			if(this.isTouching(creature)){

				// check to see if the user specified any restrictions on mating
				if(Creature.matingRestriction.equals("COMPATIBLE")){
					if(!this.brain.isCompatible(creature.brain)) continue;
				}else if(Creature.matingRestriction.equals("SAME_SPECIES")){
					if(this.speciesNum != creature.speciesNum) continue;
				}

				Creature.mateCreatures(this, creature, sim);

				this.sexualReproductions++;
				creature.sexualReproductions++;

				this.justMated = Creature.SHOW_MATING_ALERT_TIME;
				creature.justMated = Creature.SHOW_MATING_ALERT_TIME;

				// update the energy
				this.addEnergy(-Creature.MATING_ENERGY_THRESHOLD/2);
				creature.addEnergy(-Creature.MATING_ENERGY_THRESHOLD/2);

			}

		}

	}

	private void asexualReproduction(SpeciesSim sim){

		while(this.getEnergy() >= Creature.ASEXUAL_MATE_THRESHOLD){

			Creature.mateCreatures(this, this, sim);

			this.asexualReproductions++;
			this.justMated = Creature.SHOW_MATING_ALERT_TIME;

			this.addEnergy(-Creature.ASEXUAL_MATE_THRESHOLD/2);

		}

	}

	/* takes in two creatures and mates them
	 **/
	private static void mateCreatures(Creature parent1, Creature parent2, SpeciesSim sim){

		Creature childCreature;
		MutableDouble2D childPosition, childHeading;

		// obtain an initial, random heading
		while((childHeading = new MutableDouble2D(sim.random.nextDouble(), sim.random.nextDouble())).length() > 1.0) ;
	
		if(sim.random.nextBoolean()) childHeading.x *= -1.0;
		if(sim.random.nextBoolean()) childHeading.y *= -1.0;

		if(childHeading.lengthSq() > 0.0) childHeading.normalize();

		// obtain the position
		childPosition = new MutableDouble2D();

		childPosition.add(parent1.position, parent2.position);
		childPosition.multiplyIn(0.5);

		childCreature = new Creature(childPosition, childHeading, parent1.brain.createOffspring(parent2.brain));

		childCreature.setLook(Creature.recombineLooks(parent1.getLook(), parent2.getLook(), sim));
		childCreature.resetColor();

		sim.creatures.setObjectLocation(childCreature, new Double2D(childPosition));
		childCreature.stoppable = sim.schedule.scheduleRepeating(childCreature);

		childCreature.setEnergy(Creature.BABY_STARTING_ENERGY);

	}

	// recombine the looks of two parents using a non-uniform crossover method
	private static boolean[] recombineLooks(boolean[] look1, boolean[] look2, SpeciesSim sim){

		boolean[] newLook = new boolean[Entity.NUM_BITS_IN_LOOK];

		// uniform crossover
		for(int i = 0; i < Entity.NUM_BITS_IN_LOOK; i++){

			if(sim.random.nextBoolean()){
				newLook[i] = look1[i];
			}else{
				newLook[i] = look1[i];
			}

			// mutate the bit
			if(sim.random.nextDouble() < Creature.LOOK_MUT_RATE){
				newLook[i] = !newLook[i];
			}

		}

		return newLook;

	}

	public Color speciesColor(){
		return Creature.speciesColors[this.speciesNum % Creature.speciesColors.length];
	}

	public void setSpeciesNum(int n){
		this.speciesNum = n;
	}

	// changes the speed of the creature based on the value given
	private void addSpeed(double delta){

		this.setSpeed(this.getSpeed() + delta);

		if(this.getSpeed() < 0.0) this.setSpeed(0.0);
		if(this.getSpeed() > Creature.DEFAULT_SPEED) this.setSpeed(Creature.DEFAULT_SPEED);

	}

	public int getAsexualReproductions(){
		return this.asexualReproductions;
	}

	public int getSexualReproductions(){
		return this.sexualReproductions;
	}

	public int getStationaryFoodEaten(){
		return this.stationaryFoodEaten;
	}

	public int getMovingFoodEaten(){
		return this.movingFoodEaten;
	}

	public double[] getInputToBrain(){
		return this.inputToBrain(Creature.sSim);
	}


	public boolean getPrintBrain(){
		return false;
	}

	public void setPrintBrain(boolean b){

		if(b){
			System.out.println(this.brain.toString());
		}

	}

	public int getID(){
		return this.id;
	}

	public static void setVisionDivisions(int n){
		Creature.VISION_DIVISIONS = n;
	}

	public static int numVisionDivisions(){
		return Creature.VISION_DIVISIONS;
	}

	public static int numInputsToBrain(){

		if(Creature.inputToBrainType.equals("EXPLICIT")){
			return (Creature.VISION_DIVISIONS + 1) * 6; // # types of entities in simulation
		}else if(Creature.inputToBrainType.equals("EXPLICIT_SIMPLE")){
			return (Creature.VISION_DIVISIONS + 1) * 3; // # types of entities in simulation
		}else if(Creature.inputToBrainType.equals("IMPLICIT")){
			return (Creature.VISION_DIVISIONS + 1) * Entity.NUM_BITS_IN_LOOK;
		}else{
			return -1;
		}

	}

	public static int numActions(){
		return Creature.NUM_ACTIONS;
	}

	public static void setLookMutRate(double d){
		Creature.LOOK_MUT_RATE = d;
	}

	public static void setMatingEnergyThreshold(int n){
		Creature.MATING_ENERGY_THRESHOLD = n;
	}

	public static void setAsexualMatingThreshold(int n){
		Creature.ASEXUAL_MATE_THRESHOLD = n;
	}

	public static void setSharpAngleTurn(double d){
		Creature.SHARP_ANGLE_TURN = d;
	}

	public static void setWideAngleTurn(double d){
		Creature.WIDE_ANGLE_TURN = d;
	}

	public static void setMatingRestriction(String str){
		Creature.matingRestriction = str;
	}

}
