package sim.app.species;

import sim.engine.*;
import sim.portrayal.*;
import sim.util.*;
import java.util.*;

public class MovingFood extends Entity{

	private static double DEFAULT_RAND_VECT_LEN;

	private static boolean DEFAULT_SHOW_LOOK = false;
	private static double DEFAULT_DIAMETER = 75.0;
	private static double DEFAULT_SENSE_DIST = 75.0;
	private static double DEFAULT_SPEED = 10.0;
	private static boolean[] DEFAULT_LOOK = new boolean[Entity.NUM_BITS_IN_LOOK];
	private static int DEFAULT_LIFE_SPAN = 0;

	private static int energyAddedToCreature = 1000;

	public MovingFood(MutableDouble2D initPos, MutableDouble2D initHead){

		super(initPos, initHead);

		this.setDiameter(MovingFood.DEFAULT_DIAMETER);
		this.setShowLook(MovingFood.DEFAULT_SHOW_LOOK);
		this.setSenseDistance(MovingFood.DEFAULT_SENSE_DIST);
		this.setSpeed(MovingFood.DEFAULT_SPEED);
		this.setLook(MovingFood.DEFAULT_LOOK);
		this.setLifeSpan(MovingFood.DEFAULT_LIFE_SPAN);

		this.resetColor();

		// WILL CHANGE!!!!!!
		this.setEnergy(MovingFood.energyAddedToCreature);

	}

	public MovingFood(){ super(); }

	public static void setDefaults(boolean showLook, double diameter, double speed, boolean[] look, double senseDist, int lifeSpan){

		MovingFood.DEFAULT_SHOW_LOOK = showLook;
		MovingFood.DEFAULT_DIAMETER = diameter;
		MovingFood.DEFAULT_SPEED = speed;
		MovingFood.DEFAULT_LOOK = look;
		MovingFood.DEFAULT_SENSE_DIST = senseDist;
		MovingFood.DEFAULT_LIFE_SPAN = lifeSpan;

	}

	// this gives the amount of randomness added to the final vector
	public static void setRandomVectorLength(double d){

		MovingFood.DEFAULT_RAND_VECT_LEN = d;

	}

	public static void setEnergyAddedToCreature(int n){
		MovingFood.energyAddedToCreature = n;
	}

	/* simply updates the agent's position based off of its heading and current position
	 **/
	public void step(SimState state){

		SpeciesSim sim = (SpeciesSim)state;

		if(this.getAge() > this.getLifeSpan()){
			sim.removeEntityFromSim(this, sim.movingFood);
			return;
		}

		this.incrementAge();

		this.moveAwayFromCreatures(sim);

		// updates the heading of the entity if it is hitting a wall
		this.updateHeadingHitWall(sim.fieldWidth, sim.fieldHeight);

		/* update the position in the Continuous2D container
		 * also, update the position within the agent
		 **/
		this.updatePosition();
		sim.movingFood.setObjectLocation(this, new Double2D(this.position));

	}

	/* moving food tries to avoid all creatures that it can currently see
	 * this updates the MovingFood's current heading
	 **/
	private void moveAwayFromCreatures(SpeciesSim sim){

		Bag nearbyCreatures = sim.creatures.getObjectsWithinDistance(new Double2D(this.position), this.senseDistance());
		MutableDouble2D vectorAwayFromCreature = new MutableDouble2D();
		MutableDouble2D finalVector = new MutableDouble2D();
		MutableDouble2D randomVector;
		Creature creature;
		double dist;

		for(int i = 0; i < nearbyCreatures.numObjs; i++){

			creature = (Creature)nearbyCreatures.objs[i];

			dist = this.position.distance(creature.position);

			if(dist > 0.0 && dist - creature.getRadius() <= this.senseDistance()){

				vectorAwayFromCreature.subtract(this.position, creature.position);
				vectorAwayFromCreature.normalize();
	
				finalVector.addIn(vectorAwayFromCreature);

			}

		}

		this.heading = finalVector;

		if(finalVector.lengthSq() > 0.0){

			// generate the random vector
			if(MovingFood.DEFAULT_RAND_VECT_LEN > 0.0){

				while(true){
					randomVector = new MutableDouble2D(sim.random.nextDouble() * MovingFood.DEFAULT_RAND_VECT_LEN, sim.random.nextDouble() * MovingFood.DEFAULT_RAND_VECT_LEN);
	
					if(randomVector.lengthSq() > 0.0 && randomVector.length() <= MovingFood.DEFAULT_RAND_VECT_LEN) break;
	
				}

				if(sim.random.nextBoolean()) randomVector.x *= -1.0;
				if(sim.random.nextBoolean()) randomVector.y *= -1.0;

				this.heading.normalize();

				this.heading.addIn(randomVector);

			}

			this.heading.normalize();
		}

	}

}
