package sim.app.species;

import sim.engine.*;
import sim.portrayal.*;
import sim.util.*;
import java.util.*;

public class Turret extends Entity{

	private static int SHOOT_INTERVAL = 50;

	private static boolean DEFAULT_SHOW_LOOK = false;
	private static double DEFAULT_DIAMETER = 75.0;
	private static double DEFAULT_SENSE_DIST = 75.0;
	private static double DEFAULT_SPEED = 10.0;
	private static boolean[] DEFAULT_LOOK = new boolean[Entity.NUM_BITS_IN_LOOK];
	private static int DEFAULT_LIFE_SPAN = 0;

	private int timeSinceShot;

	public Turret(MutableDouble2D initPos, MutableDouble2D initHead){

		super(initPos, initHead);

		this.setDiameter(Turret.DEFAULT_DIAMETER);
		this.setShowLook(Turret.DEFAULT_SHOW_LOOK);
		this.setSenseDistance(Turret.DEFAULT_SENSE_DIST);
		this.setSpeed(Turret.DEFAULT_SPEED);
		this.setLook(Turret.DEFAULT_LOOK);
		this.setLifeSpan(Turret.DEFAULT_LIFE_SPAN);

		this.resetColor();

		// WILL CHANGE!!!!!!
		this.setEnergy(1000);

		this.timeSinceShot = 0;

	}

	public Turret(){ super(); }

	public static void setDefaults(boolean showLook, double diameter, double speed, boolean[] look, double senseDist, int lifeSpan){

		Turret.DEFAULT_SHOW_LOOK = showLook;
		Turret.DEFAULT_DIAMETER = diameter;
		Turret.DEFAULT_SPEED = speed;
		Turret.DEFAULT_LOOK = look;
		Turret.DEFAULT_SENSE_DIST = senseDist;
		Turret.DEFAULT_LIFE_SPAN = lifeSpan;

	}

	public static void setShootInterval(int n){

		Turret.SHOOT_INTERVAL = n;

	}

	/* simply updates the agent's position based off of its heading and current position
	 **/
	public void step(SimState state){

		SpeciesSim sim = (SpeciesSim)state;

		if(this.timeSinceShot >= Turret.SHOOT_INTERVAL){
			this.shootBullet(sim);
			this.timeSinceShot = 0;
		}else{
			this.timeSinceShot++;
		}

	}

	private void shootBullet(SpeciesSim sim){

		Bullet newBullet;

		newBullet = new Bullet(new MutableDouble2D(this.position), new MutableDouble2D(this.heading));

		sim.bullets.setObjectLocation(newBullet, new Double2D(this.position));
		newBullet.stoppable = sim.schedule.scheduleRepeating(newBullet);

	}

}
